;============================================================================
;	STM_PORTB.ASM
; 
;	Simple "driver" for LEDs and stepper motor on Port B.
;	LED control via macro set_led;
;	Stepper motor control uses timer interrupt.
;
;	Copyright (C) Jrgen Mller, juergen@e-basteln.de
;
;	Port B, bits 0..6	stepper motor driver, TCA 3727
;	Port B, bit  7		motor limit switch, input (0 = lower limit reached)
;	Port B, bits 8 ..11	2 triple-color LEDs
;	Port B, bits 12,14	test outputs
;
;	jm  6.11.99	adapted to STM_MAIN environment
;	jm 14.11.99	uses cmd_md, cmd_mp variables
;	jm 27.11.99	debugging
;	jm 28.11.99	introduced non-chopped stepper operation 
;			(to reduce coupling to tunneling current amplifier)
;	jm 5.1.2000	moved approach and retract cmd handlers to stm_ssi;
;			introduced wait state before movement to allow for
;			piezo to safely retract before approach/retract motions
;	jm 6.1.2000	renamed to STM_PORTB
;	jm 8.1.2000	test outputs on pin 12,14 added
;	
;	Automatic power-off after each move is not stable while motor under load!
;	(Neither for 1/2 nor 1/5 step mode.) Power should be kept on as a default;
;	explicit power-off supported via motor_on, motor_off macros.
;	(May introduce small motor movement!)
;
;	Chopped motor supply (current regulation) disturbs the tunneling current
;	pre-amp and should not be used.
;
;============================================================================

STEP5		EQU	0		; 1 = 1/5 steps, 0 = 1/2 steps
CHOPPED		EQU	0		; 1 = use reduces power levels (chopped supply)
POWEROFF	EQU	0		; 1 = motor power off after motor moves
FREQ		EQU	100		; stepping frequency in Hz

;----------------------------------------------------------------------------
;	LED control
;----------------------------------------------------------------------------

; LED control
; Usage: set_led  led_number(0=right, 1=left),color(0..3)

set_led	macro	led,color
	if color&1
	  bset	#(8+2*led),x:PBD
	else
	  bclr	#(8+2*led),x:PBD
	endif
	if color&2
	  bset	#(9+2*led),x:PBD
	else
	  bclr	#(9+2*led),x:PBD
	endif
	endm

; LED colors for use with set_led

off	equ	0
red	equ	1
green	equ	2
yellow	equ	3

; LED positions

led_motor	equ	1		; motor indicator (left)
led_tunnel	equ	0		; tunneling indicator (right)

;----------------------------------------------------------------------------
;	Test output control
;----------------------------------------------------------------------------

; Test output control
; Usage: set_test  pin_number(0=12, 1=14),state (0,1)

set_test macro	pin,state
	if state
	  bset	#(12+2*pin),x:PBD
	else
	  bclr	#(12+2*pin),x:PBD
	endif
	endm

toggle_test macro pin
	bchg	#(12+2*pin),x:PBD
	endm

;----------------------------------------------------------------------------
;	Stepper motor control macros
;----------------------------------------------------------------------------

motor_limit	macro			; return CC when limit switch closed
	btst	#mot_switch,x:PBD
	endm

motor_off	macro			; disable motor driver
	bclr	#mot_enable,x:PBD
	endm

motor_on	macro			; enable motor driver
	bset	#mot_enable,x:PBD
	endm

motor_zero	macro			; define current position as 0
	move	#0,r0
	move	r0,x:cmd_mp
	endm

motor_status	macro			; return CS when motor ready
	btst	#mode_done,x:mode
	endm

motor_up1	macro			; move up 1 step
	move	#(1<<mode_single)+(1<<mode_wait),r0
	move	r0,x:mode
	endm

motor_down1	macro			; move down 1 step
	move	#(1<<mode_single)+(1<<mode_down)+(1<<mode_wait),r0
	move	r0,x:mode
	endm

motor_up	macro			; move up to cmd_md
	move	#(1<<mode_wait),r0
	move	r0,x:mode
	endm

motor_down	macro			; move down to cmd_md
	move	#(1<<mode_down)+(1<<mode_wait),r0
	move	r0,x:mode
	endm

;----------------------------------------------------------------------------
;	interrupt vectors
;----------------------------------------------------------------------------

	org     p:$003C
	jsr     timer_isr		;Timer interrupt

;----------------------------------------------------------------------------
;	constants
;----------------------------------------------------------------------------

; stepper motor phase settings

p0		equ	%1000011	; phase P, current values 0..3
p1		equ	%1000010	; table includes bit6 = enable!
p2		equ	%1000001
p3		equ	%1000000
mp		equ	%0000100	; reverse current

q0		equ	%0011000	; phase Q, current values 0..3
q1		equ	%0010000
q2		equ	%0001000
q3		equ	%0000000
mq		equ	%0100000	; reverse current

mot_switch	equ	7		; no. of switch input bit
mot_enable	equ	6		; no. of enable bit

;----------------------------------------------------------------------------
;	variables
;----------------------------------------------------------------------------
       		org    x:

;--------------------------------------------------------------------
; stepper motor stepping table, 1/5 steps (1000/rev)

		IF	STEP5

		buffer	M,32
steps		dc	p3+q1
		dc	p3+q2
		dc	p2+q3
		dc	p1+q3
		dc	p0+q3
		dc	p1+q3+mp
		dc	p2+q3+mp
		dc	p3+q2+mp
		dc	p3+q1+mp
		dc	p3+q0+mp
		dc	p3+q1+mp+mq
		dc	p3+q2+mp+mq
		dc	p2+q3+mp+mq
		dc	p1+q3+mp+mq
		dc	p0+q3+mp+mq
		dc	p1+q3+mq
		dc	p2+q3+mq
		dc	p3+q2+mq
		dc	p3+q1+mq
		dc	p3+q0+mq

countmax	equ	*-steps-1	; length of table-1
		endbuf

		ELSE			; not step5

;--------------------------------------------------------------------
; stepper motor stepping table, 1/2 steps (400/rev)

		IF 	CHOPPED
		buffer	M,8
steps		dc	p2+q2
		dc	p0+q3
		dc	p2+q2+mp
		dc	p3+q0+mp
		dc	p2+q2+mp+mq
		dc	p0+q3+mp+mq
		dc	p2+q2+mq
		dc	p3+q0+mq

countmax	equ	*-steps-1	; length of table-1
		endbuf

;--------------------------------------------------------------------
; stepper motor stepping table, 1/2 steps (400/rev)
; only full-power setting used to avoid chopping

		ELSE			; not chopped
		buffer	M,8
steps		dc	p3+q3
		dc	p0+q3
		dc	p3+q3+mp
		dc	p3+q0+mp
		dc	p3+q3+mp+mq
		dc	p0+q3+mp+mq
		dc	p3+q3+mq
		dc	p3+q0+mq

countmax	equ	*-steps-1	; length of table-1
		endbuf

		ENDIF			; chopped?
		ENDIF			; step5?

; stepper motor state variables
		org	x:
counter		dc      0       ; step (state) counter, 0..countmax
mode		dc	0	; bit set, controls timer irq action (see def's below)

; bits in stepper motor mode word 
mode_down	equ	0	; direction reversed 
mode_cont	equ	1	; run continuously until limit switch reached
mode_dest	equ	2	; run continuously until destination reached
mode_single	equ	3	; single step, then power off
mode_off	equ	4	; power off in next interrupt
mode_wait	equ	5	; wait one timer tick before beginning movement
mode_done	equ	15	; output: movement complete

;---------------------------------------------------------------------------
;
;      DSP initialization
;
;---------------------------------------------------------------------------
	org     ph:			;"slow" code

portb_init
	; initialize port B
	movep	#0,x:PBC		; port B is general purpose
	movep	#$FF7F,x:PBDDR		; H0-H15 are outputs (H7 is input)

	; initialize stepping sequence
	move	#steps,r0
	move	r0,x:counter
	move	#1<<mode_done,r0
	move	r0,x:mode

	; set both LEDs to green
	set_led	0,green
	set_led	1,green

	; initialize timer
	movep	#(30000000/FREQ-1),x:TCR ;timer period (CLK/2 = 30 MHz) -> 300 Hz
	movep	#$0003,x:TCSR		;mode 0, enable timer & interrupt

	rts

;--------------------------------------------------------------------
;   Motor commands 
;--------------------------------------------------------------------

mot_goto
	; goto cmd_md position

	set_led	led_motor,yellow
	move	x:cmd_md,a
	move	x:cmd_mp,x0
	sub	x0,a
	jcs	<_down
	motor_up
	jmp	<_return
_down	motor_down
_return	rts

mot_limit
	; move down to limit switch,
	; define limit pos. as 0

	set_led	led_motor,yellow
	move	#-1,x0
	move	x0,x:cmd_mp		; assume max. current position 
	move	#0,x0
	move	x0,x:cmd_md		; destination is 0 
	motor_down			; move way down
	rts

mot_stop
	; stop motor immediately

	bset	#mode_done,x:mode	; no more IRQ action
	bclr	#mot_enable,x:PBD	; disable stepper motor power
	set_led	led_motor,green		; indicate stop state
	rts

mot_plus
	; one step up
	motor_up1
	rts
	
mot_minus
	; one step down
	motor_down1
	rts

mot_off
	; switch power off
	motor_off
	rts

mot_on
	; switch power on
	motor_on
	rts

;--------------------------------------------------------------------
;   Timer Interrupt Service Routine 
;   R6 should point to the first free location on the stack (in Y:memory).
;--------------------------------------------------------------------

	org	pl:			; "fast" code

timer_isr
	move    r0,y:(r6)+              ; Save r0 to the stack.
	move    m0,y:(r6)+              ; Save m0 to the stack.
	move	a0,y:(r6)+
	move	a1,y:(r6)+
	move	a2,y:(r6)+
	move	x0,y:(r6)+
	move	x1,y:(r6)+
	jset	#mode_done,x:mode,_return ; nothing to do

	bclr	#mode_wait,x:mode	; wait before starting?
	jcs	<_return		; yes

	bclr	#mode_off,x:mode	; ready to power off?
	jcs	<_off			; yes

	; prepare to move 1 step...
	move	x:cmd_mp,a	
	move	#>1,x0
	move	x:counter,r0		; get current motor state
	move	#countmax,m0		; mod x addressing
	jset	#mode_down,x:mode,_down	; move forward or back?

	; move up
	add	x0,a
	move	x:-(r0),x0
	jmp	<_cont

	; move down
_down	motor_limit			; test limit switch
	jcc	<_limit			; reached
	sub	x0,a
	move	x:(r0)+,x0

_cont	move	r0,x:counter		; update stepper state and absolute position
	move	a,x:cmd_mp

	move	x:PBD,a			; move stepper state to PBD, leave LED state intact
	move	#$ff00,x1
	and	x1,a1
	add	x0,a
	move	a1,x:PBD

	jset	#mode_single,x:mode,_last ; single step: power off next
	move	x:cmd_md,x0
	move	x:cmd_mp,a
	cmp	x0,a			; destination reached?
	jeq	<_last			; yes
	jmp	<_return		; no

_limit	move	#0,x0			; limit switch reached: define current position as 0
	move	x0,x:cmd_md
	move	x0,x:cmd_mp

_last	bset	#mode_off,x:mode	; power off in next cycle
	jmp	<_return

_off	
	IF	POWEROFF
	mot_off				; disable stepper motor power
	ENDIF

_done	bset	#mode_done,x:mode
	set_led	led_motor,green		; indicate stop state

_return	move    y:-(r6),x1
	move    y:-(r6),x0
	move    y:-(r6),a2
	move    y:-(r6),a1
	move    y:-(r6),a0
	move    y:-(r6),m0
	move    y:-(r6),r0

	rti
