;============================================================================
;	STM_SCI.ASM
; 
;	Driver for serial interface, with ring buffers for RX and TX.
;	Copyright (C) Jrgen Mller, juergen@e-basteln.de
;
;	RS-232 mode: 15200 Baud, 8 data bits, 1 stop bit, no parity
;
;	jm 9.11.98
;	jm 29.9.99	modified for 60 MHz clock, 19200 baud 
;	jm 6.11.99	adapted to STM_MAIN environment
;	jm 7.11.99	bug-fixing, added CR counting
;	jm 8.11.99	removed local IPR setting
;	jm 5.1.2000	changed to 115 kBaud
;
;============================================================================

;----------------------------------------------------------------------------
;	interrupt vectors
;----------------------------------------------------------------------------

	org     p:$0014

	jsr     sci_rx			; SCI receive interrupt
	jsr     sci_rx			; SCI receive interrupt
	jsr	sci_tx			; SCI transmitter empty
	jsr	sci_tx			; SCI transmitter empty

;----------------------------------------------------------------------------
;	variables
;----------------------------------------------------------------------------

		org	x:		; "fast" variables

rx_rdptr	ds	1		; Read pointer for rx buffer
rx_wrptr	ds	1		; Write pointer for rx buffer
tx_rdptr	ds	1		; Read pointer for tx buffer
tx_wrptr	ds	1		; Write pointer for tx buffer
rx_crcount	ds	1		; number of CRs in buffer

		org	y:		; "slow" variables

SCI_RX_BUFLEN	equ	$100
SCI_TX_BUFLEN	equ	$2000

sci_rx_buf	dsm	SCI_RX_BUFLEN
sci_tx_buf	dsm	SCI_TX_BUFLEN


;----------------------------------------------------------------------------
;	SCI constants
;----------------------------------------------------------------------------

BEL		equ	$07		; ASCII BEL
CR		equ	$0D		; ASCII CR (marks end of line)

RIE		equ	11		; rx int enable (bit pos in SCR)
TIE		equ	12		; tx int enable (bit pos in SCR)

;---------------------------------------------------------------------------
;      DSP initialization
;---------------------------------------------------------------------------
	org     ph:

sci_init
	; initialize SCI
	movep	#$0302,x:SCR		; enable transmitter, receiver
					; point to point, 8n1 (10 bit async)
					; SSI interrupts remain disabled here.
;	movep	#$0030,x:SCCR		; 19200 bps, internal clock
	movep	#$0007,x:SCCR		; 115200 bps, internal clock
					; divider = 60 MHz/(4*16*baud rate) - 1
	bset	#0,x:PCC		; select RX, TX for SCI
	bset	#1,x:PCC		; (PCC partially initialized by SSI!)

	jmp	sci_flush		; initialize read/write buffers

;--------------------------------------------------------------------
;   SCI Flush Subroutine
;   Flushes (initializes) rx and tx buffers
;
;   Modifies X0
;--------------------------------------------------------------------

sci_flush
	bclr	#RIE,x:SCR
	bclr	#TIE,x:SCR
	move	#>sci_rx_buf,x0
	move	x0,x:rx_rdptr
	move	x0,x:rx_wrptr
	move	#>sci_tx_buf,x0
	move	x0,x:tx_rdptr
	move	x0,x:tx_wrptr
	move	#0,x0
	move	x0,x:rx_crcount
	bset	#RIE,x:SCR		; enable rx interrupt, tx remains disabled
	rts

;--------------------------------------------------------------------
;   SCI Receive Subroutine
;   Returns character from RX buffer in A1
;   Returns A=0, EQ flag if buffer is empty
;   Modifies R1,M1,X0,A
;--------------------------------------------------------------------

sci_read
	bclr	#RIE,x:SCR		; disable rx interrupt
	move    #(SCI_RX_BUFLEN-1),m1	; Modulo BUFLEN buffer.
	move	x:rx_rdptr,r1   	; Load the read pointer to the rx buffer.
	move	x:rx_wrptr,x0

	; check whether buffer contains data
	move	r1,a	
	sub	x0,a			; characters left in buffer?
	jeq	<_empty	

	; read character into a1
	move	y:(r1)+,a         	; Read out received data to buffer.
	move    r1,x:rx_rdptr		; Update rx buffer pointer.

	; check for CR, update count
	move	#>CR,x0
	cmp	x0,a
	jne	<_empty
	move	x:rx_crcount,a0
	dec	a
	move	a0,x:rx_crcount
	move	x0,a			; restore CR
	tst	a			; set NE flag!

_empty	bset	#RIE,x:SCR		; enable rx interrupt
	rts

;--------------------------------------------------------------------
;   SCI Transmit Subroutine
;   A1 contains data to be put into TX buffer 
;   Modifies R1,M1,X0,A
;--------------------------------------------------------------------

sci_write
	bclr	#TIE,x:SCR		; disable tx interrupt
	move    #(SCI_TX_BUFLEN-1),m1	; Modulo BUFLEN buffer.
	move    x:tx_wrptr,r1           ; Load the pointer to the tx buffer.
	move	x:tx_rdptr,x0		; load read pointer
	move    a1,y:(r1)+
	move	r1,x:tx_wrptr           ; Update tx buffer pointer.

	move    r1,a	
	sub	x0,a
	jne	<_noovfl		; overflow?
	move	x0,r1            
	nop
	move	(r1)+
	move    r1,x:tx_rdptr           ; Update tx read buffer pointer.
_noovfl	bset	#TIE,x:SCR		; enable tx interrupt
	rts

; Put character into serial output buffer

putchar	macro	char
	move	char,a1
	jsr	sci_write
	endm

;--------------------------------------------------------------------
;   SCI Receive ISR
;   R6 should point to the first free location on the stack (in Y:memory).
;--------------------------------------------------------------------

	org	pl:			; "fast" code
sci_rx
	move	r0,y:(r6)+              ; Save r0 to the stack.
	move	m0,y:(r6)+              ; Save m0 to the stack.
	move	a0,y:(r6)+
	move	a1,y:(r6)+
	move	a2,y:(r6)+
	move	x0,y:(r6)+

	; store received byte in buffer
	move    #(SCI_RX_BUFLEN-1),m0	; Modulo BUFLEN buffer.
	move    x:rx_wrptr,r0           ; Load the write pointer to the rx buffer.
	movep   x:SRX,x0
	move	x0,y:(r0)+         	; Read out received data to buffer.
	move    r0,x:rx_wrptr		; Update rx buffer pointer.

	; count CRs
	clr	a
	move	#CR,a1
	cmp	x0,a
	jne	<_nocr
	move	x:rx_crcount,a0
	inc	a
	move	a0,x:rx_crcount
	

	; check for overflows; on overflow discard oldest character
_nocr	move	x:rx_rdptr,x0
	move	r0,a
	sub	x0,a
	jne	<_noovfl		; space left in buffer?
	move	x0,r0           	; overflow: discard oldest character
	nop
	move	(r0)+	
	move	r0,x:rx_rdptr

_noovfl	move	y:-(r6),x0
	move	y:-(r6),a2
	move	y:-(r6),a1
	move	y:-(r6),a0
	move	y:-(r6),m0
	move	y:-(r6),r0
	rti

;--------------------------------------------------------------------
;   SCI Transmit ISR
;   R6 should point to the first free location on the stack (in Y:memory).
;--------------------------------------------------------------------

	org	pl:			; "fast" code
sci_tx
	move    r0,y:(r6)+
	move    m0,y:(r6)+
	move	a0,y:(r6)+
	move	a1,y:(r6)+
	move	a2,y:(r6)+
	move	x0,y:(r6)+

	move    #(SCI_TX_BUFLEN-1),m0	; Modulo BUFLEN buffer.
	move    x:tx_rdptr,r0		; Load the read pointer to the tx buffer.
	move	x:tx_wrptr,x0
	movep   y:(r0)+,x:STX         	; SSI transfer data register.
	move    r0,x:tx_rdptr           ; Update tx buffer pointer.

	move	r0,a
	sub	x0,a
	jne	<_more			; all transmitted?
	bclr	#TIE,x:SCR		; yes - stop tx interrupts

_more	move    y:-(r6),x0
	move    y:-(r6),a2
	move    y:-(r6),a1
	move    y:-(r6),a0
	move    y:-(r6),m0
	move    y:-(r6),r0
	rti

;--------------------------------------------------------------------
;   SCI transmit with "compression":
;   Transmit 16 bit value in A1. 
;   Transmit difference from last value if it is in the range +-127.
;   Otherwise, insert flag -128 and absolute 16 bit value.
;
;   Modifies R1,M1,X0,A
;   R6 should point to the first free location on the stack (in Y:memory).
;--------------------------------------------------------------------

sci_tx16
	
